#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.parsing import graken, Parser
from grako.util import re, RE_FLAGS


__version__ = (2015, 12, 26, 22, 15, 59, 5)

__all__ = [
    'BParser',
    'BSemantics',
    'main'
]


class BParser(Parser):
    def __init__(self,
                 whitespace=None,
                 nameguard=None,
                 comments_re='/\\*((?:[^\\*]|\\*[^/]|\\n)*?)\\*+/',
                 eol_comments_re=None,
                 ignorecase=None,
                 left_recursion=False,
                 **kwargs):
        super(BParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            **kwargs
        )

    @graken()
    def _program_(self):

        def block1():
            self._definition_()
            self._cut()
        self._closure(block1)
        self.ast['@'] = self.last_node
        self._check_eof()

    @graken()
    def _definition_(self):
        with self._choice():
            with self._option():
                self._simpledef_()
            with self._option():
                self._vectordef_()
            with self._option():
                self._functiondef_()
            self._error('no available options')

    @graken()
    def _simpledef_(self):
        self._name_()
        self.ast['name'] = self.last_node
        with self._optional():
            self._ival_()
        self.ast['init'] = self.last_node
        self._token(';')

        self.ast._define(
            ['name', 'init'],
            []
        )

    @graken()
    def _vectordef_(self):
        self._name_()
        self.ast['name'] = self.last_node
        self._token('[')
        with self._optional():
            self._constantexpr_()
            self.ast['maxidx'] = self.last_node
        self._token(']')
        with self._optional():
            self._ivallist_()
            self.ast['ivals'] = self.last_node
        self._token(';')

        self.ast._define(
            ['name', 'maxidx', 'ivals'],
            []
        )

    @graken()
    def _ivallist_(self):
        self._ival_()
        self.ast.setlist('@', self.last_node)

        def block1():
            self._token(',')
            self._ival_()
            self.ast.setlist('@', self.last_node)
        self._closure(block1)

    @graken()
    def _ival_(self):
        with self._choice():
            with self._option():
                self._numericexpr_()
            with self._option():
                self._characterexpr_()
            with self._option():
                self._stringexpr_()
            self._error('no available options')

    @graken()
    def _functiondef_(self):
        self._name_()
        self.ast['name'] = self.last_node
        self._token('(')
        with self._optional():
            self._namelist_()
            self.ast['args'] = self.last_node
        self._token(')')
        self._cut()
        self._statement_()
        self.ast['body'] = self.last_node

        self.ast._define(
            ['name', 'args', 'body'],
            []
        )

    @graken()
    def _statement_(self):
        with self._choice():
            with self._option():
                self._labelstatement_()
            with self._option():
                self._gotostatement_()
            with self._option():
                self._switchstatement_()
            with self._option():
                self._casestatement_()
            with self._option():
                self._breakstatement_()
            with self._option():
                self._autostatement_()
            with self._option():
                self._extrnstatement_()
            with self._option():
                self._compoundstatement_()
            with self._option():
                self._ifstatement_()
            with self._option():
                self._whilestatement_()
            with self._option():
                self._returnstatement_()
            with self._option():
                self._exprstatement_()
            with self._option():
                self._nullstatement_()
            self._error('no available options')

    @graken()
    def _labelstatement_(self):
        with self._ifnot():
            with self._group():
                self._token('default')
        self._name_()
        self.ast['label'] = self.last_node
        self._token(':')
        self._statement_()
        self.ast['statement'] = self.last_node

        self.ast._define(
            ['label', 'statement'],
            []
        )

    @graken()
    def _gotostatement_(self):
        self._token('goto')
        self._cut()
        self._name_()
        self.ast['label'] = self.last_node
        self._token(';')

        self.ast._define(
            ['label'],
            []
        )

    @graken()
    def _switchstatement_(self):
        self._token('switch')
        self._cut()
        self._expr_()
        self.ast['rvalue'] = self.last_node
        self._cut()
        self._statement_()
        self.ast['body'] = self.last_node

        self.ast._define(
            ['rvalue', 'body'],
            []
        )

    @graken()
    def _casestatement_(self):
        with self._group():
            with self._choice():
                with self._option():
                    with self._group():
                        self._token('case')
                        self._constantexpr_()
                        self.ast['cond'] = self.last_node
                with self._option():
                    self._token('default')
                self._error('expecting one of: default')
        self._cut()
        self._token(':')
        self._statement_()
        self.ast['then'] = self.last_node

        self.ast._define(
            ['cond', 'then'],
            []
        )

    @graken()
    def _breakstatement_(self):
        self._token('break')
        self._token(';')

    @graken()
    def _autostatement_(self):
        self._token('auto')
        self._cut()
        self._autovar_()
        self.ast.setlist('@', self.last_node)

        def block1():
            self._token(',')
            self._autovar_()
            self.ast.setlist('@', self.last_node)
        self._closure(block1)
        self._token(';')

    @graken()
    def _autovar_(self):
        self._name_()
        self.ast['name'] = self.last_node
        with self._optional():
            self._token('[')
            self._constantexpr_()
            self.ast['maxidx'] = self.last_node
            self._token(']')

        self.ast._define(
            ['name', 'maxidx'],
            []
        )

    @graken()
    def _extrnstatement_(self):
        self._token('extrn')
        self._cut()
        self._namelist_()
        self.ast['@'] = self.last_node
        self._token(';')

    @graken()
    def _compoundstatement_(self):
        self._token('{')
        self._cut()

        def block1():
            self._statement_()
            self._cut()
        self._closure(block1)
        self.ast['@'] = self.last_node
        self._token('}')

    @graken()
    def _ifstatement_(self):
        self._token('if')
        self._cut()
        self._token('(')
        self._expr_()
        self.ast['cond'] = self.last_node
        self._token(')')
        self._statement_()
        self.ast['then'] = self.last_node
        with self._optional():
            self._token('else')
            self._statement_()
            self.ast['otherwise'] = self.last_node

        self.ast._define(
            ['cond', 'then', 'otherwise'],
            []
        )

    @graken()
    def _whilestatement_(self):
        self._token('while')
        self._cut()
        self._token('(')
        self._expr_()
        self.ast['cond'] = self.last_node
        self._token(')')
        self._statement_()
        self.ast['body'] = self.last_node

        self.ast._define(
            ['cond', 'body'],
            []
        )

    @graken()
    def _returnstatement_(self):
        self._token('return')
        self._cut()
        with self._optional():
            self._token('(')
            self._expr_()
            self.ast['return_value'] = self.last_node
            self._token(')')
        self._token(';')

        self.ast._define(
            ['return_value'],
            []
        )

    @graken()
    def _exprstatement_(self):
        self._expr_()
        self.ast['@'] = self.last_node
        self._token(';')

    @graken()
    def _nullstatement_(self):
        self._token(';')

    @graken()
    def _expr_(self):
        self._assignexpr_()

    @graken()
    def _assignexpr_(self):
        self._condexpr_()
        self.ast['lhs'] = self.last_node
        with self._optional():
            self._assignop_()
            self.ast['op'] = self.last_node
            self._assignexpr_()
            self.ast['rhs'] = self.last_node

        self.ast._define(
            ['lhs', 'op', 'rhs'],
            []
        )

    @graken()
    def _assignop_(self):
        self._pattern(r'=([+\-/\*%&^|]|[=!]=|>[=>]?|<[=<]?)?')

    @graken()
    def _condexpr_(self):
        self._orexpr_()
        self.ast['cond'] = self.last_node
        with self._optional():
            self._token('?')
            self._condexpr_()
            self.ast['then'] = self.last_node
            self._token(':')
            self._condexpr_()
            self.ast['otherwise'] = self.last_node

        self.ast._define(
            ['cond', 'then', 'otherwise'],
            []
        )

    @graken()
    def _orexpr_(self):
        self._xorexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._ortail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _ortail_(self):
        self._token('|')
        self.ast['op'] = self.last_node
        self._xorexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _xorexpr_(self):
        self._andexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._xortail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _xortail_(self):
        self._token('^')
        self.ast['op'] = self.last_node
        self._andexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _andexpr_(self):
        self._eqexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._andtail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _andtail_(self):
        self._token('&')
        self.ast['op'] = self.last_node
        self._eqexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _eqexpr_(self):
        self._relexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._eqtail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _eqtail_(self):
        self._eqop_()
        self.ast['op'] = self.last_node
        self._relexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _eqop_(self):
        self._pattern(r'[!=]=')

    @graken()
    def _relexpr_(self):
        self._shiftexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._reltail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _reltail_(self):
        self._relop_()
        self.ast['op'] = self.last_node
        self._shiftexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _relop_(self):
        self._pattern(r'[<>]={0,1}')

    @graken()
    def _shiftexpr_(self):
        self._addexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._shifttail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _shifttail_(self):
        self._shiftop_()
        self.ast['op'] = self.last_node
        self._addexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _shiftop_(self):
        self._pattern(r'<<|>>')

    @graken()
    def _addexpr_(self):
        self._multexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._addtail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _addtail_(self):
        self._addop_()
        self.ast['op'] = self.last_node
        self._multexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _addop_(self):
        self._pattern(r'[+-]')

    @graken()
    def _multexpr_(self):
        self._unaryexpr_()
        self.ast['lhs'] = self.last_node

        def block2():
            self._multtail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['lhs', 'tail'],
            []
        )

    @graken()
    def _multtail_(self):
        self._multop_()
        self.ast['op'] = self.last_node
        self._unaryexpr_()
        self.ast['rhs'] = self.last_node

        self.ast._define(
            ['op', 'rhs'],
            []
        )

    @graken()
    def _multop_(self):
        self._pattern(r'[/%\*]')

    @graken()
    def _unaryexpr_(self):

        def block1():
            self._leftunaryop_()
        self._closure(block1)
        self.ast['leftops'] = self.last_node
        self._primaryexpr_()
        self.ast['rhs'] = self.last_node

        def block4():
            self._rightunaryop_()
        self._closure(block4)
        self.ast['rightops'] = self.last_node

        self.ast._define(
            ['leftops', 'rhs', 'rightops'],
            []
        )

    @graken()
    def _leftunaryop_(self):
        self._pattern(r'[\*&!\~]|--?|\+\+')

    @graken()
    def _rightunaryop_(self):
        with self._choice():
            with self._option():
                self._token('++')
            with self._option():
                self._token('--')
            self._error('expecting one of: ++ --')

    @graken()
    def _primaryexpr_(self):
        self._primaryexprhead_()
        self.ast['head'] = self.last_node

        def block2():
            self._primaryexprtail_()
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['head', 'tail'],
            []
        )

    @graken()
    def _primaryexprhead_(self):
        with self._choice():
            with self._option():
                self._token('(')
                self._expr_()
                self.ast['@'] = self.last_node
                self._token(')')
            with self._option():
                self._builtinexpr_()
            with self._option():
                self._variableexpr_()
            with self._option():
                self._constantexpr_()
            with self._option():
                self._stringexpr_()
            self._error('no available options')

    @graken()
    def _primaryexprtail_(self):
        with self._choice():
            with self._option():
                self._token('(')
                with self._optional():
                    self._exprlist_()
                    self.ast['args'] = self.last_node
                self._token(')')
            with self._option():
                self._token('[')
                self._expr_()
                self.ast['index'] = self.last_node
                self._token(']')
            self._error('expecting one of: (')

        self.ast._define(
            ['args', 'index'],
            []
        )

    @graken()
    def _variableexpr_(self):
        with self._ifnot():
            self._builtinexpr_()
        self._name_()

    @graken()
    def _constantexpr_(self):
        with self._choice():
            with self._option():
                self._numericexpr_()
            with self._option():
                self._characterexpr_()
            self._error('no available options')

    @graken()
    def _builtinexpr_(self):
        self._token('__bytes_per_word')

    @graken()
    def _numericexpr_(self):

        def block0():
            self._NUMERIC_()
        self._positive_closure(block0)

    @graken()
    def _characterexpr_(self):
        self._token("'")

        def block1():
            self._CHARACTERCONSTCHAR_()
        self._closure(block1)
        self.ast['@'] = self.last_node
        self._token("'")

    @graken()
    def _stringexpr_(self):
        self._token('"')

        def block1():
            self._STRINGCONSTCHAR_()
        self._closure(block1)
        self.ast['@'] = self.last_node
        self._token('"')

    @graken()
    def _name_(self):
        self._ALPHA_()
        self.ast['head'] = self.last_node

        def block2():
            with self._choice():
                with self._option():
                    self._ALPHA_()
                with self._option():
                    self._NUMERIC_()
                self._error('no available options')
        self._closure(block2)
        self.ast['tail'] = self.last_node

        self.ast._define(
            ['head', 'tail'],
            []
        )

    @graken()
    def _ALPHA_(self):
        self._pattern(r'[A-Za-z_\.\b]')

    @graken()
    def _NUMERIC_(self):
        self._pattern(r'[0-9]')

    @graken()
    def _CHARACTERCONSTCHAR_(self):
        self._pattern(r"([^'\*])|(\*.)")

    @graken()
    def _STRINGCONSTCHAR_(self):
        self._pattern(r'([^"\*])|(\*.)')

    @graken()
    def _exprlist_(self):
        self._expr_()
        self.ast.setlist('@', self.last_node)

        def block1():
            self._token(',')
            self._expr_()
            self.ast.setlist('@', self.last_node)
        self._closure(block1)

    @graken()
    def _namelist_(self):
        self._name_()
        self.ast.setlist('@', self.last_node)

        def block1():
            self._token(',')
            self._name_()
            self.ast.setlist('@', self.last_node)
        self._closure(block1)


class BSemantics(object):
    def program(self, ast):
        return ast

    def definition(self, ast):
        return ast

    def simpledef(self, ast):
        return ast

    def vectordef(self, ast):
        return ast

    def ivallist(self, ast):
        return ast

    def ival(self, ast):
        return ast

    def functiondef(self, ast):
        return ast

    def statement(self, ast):
        return ast

    def labelstatement(self, ast):
        return ast

    def gotostatement(self, ast):
        return ast

    def switchstatement(self, ast):
        return ast

    def casestatement(self, ast):
        return ast

    def breakstatement(self, ast):
        return ast

    def autostatement(self, ast):
        return ast

    def autovar(self, ast):
        return ast

    def extrnstatement(self, ast):
        return ast

    def compoundstatement(self, ast):
        return ast

    def ifstatement(self, ast):
        return ast

    def whilestatement(self, ast):
        return ast

    def returnstatement(self, ast):
        return ast

    def exprstatement(self, ast):
        return ast

    def nullstatement(self, ast):
        return ast

    def expr(self, ast):
        return ast

    def assignexpr(self, ast):
        return ast

    def assignop(self, ast):
        return ast

    def condexpr(self, ast):
        return ast

    def orexpr(self, ast):
        return ast

    def ortail(self, ast):
        return ast

    def xorexpr(self, ast):
        return ast

    def xortail(self, ast):
        return ast

    def andexpr(self, ast):
        return ast

    def andtail(self, ast):
        return ast

    def eqexpr(self, ast):
        return ast

    def eqtail(self, ast):
        return ast

    def eqop(self, ast):
        return ast

    def relexpr(self, ast):
        return ast

    def reltail(self, ast):
        return ast

    def relop(self, ast):
        return ast

    def shiftexpr(self, ast):
        return ast

    def shifttail(self, ast):
        return ast

    def shiftop(self, ast):
        return ast

    def addexpr(self, ast):
        return ast

    def addtail(self, ast):
        return ast

    def addop(self, ast):
        return ast

    def multexpr(self, ast):
        return ast

    def multtail(self, ast):
        return ast

    def multop(self, ast):
        return ast

    def unaryexpr(self, ast):
        return ast

    def leftunaryop(self, ast):
        return ast

    def rightunaryop(self, ast):
        return ast

    def primaryexpr(self, ast):
        return ast

    def primaryexprhead(self, ast):
        return ast

    def primaryexprtail(self, ast):
        return ast

    def variableexpr(self, ast):
        return ast

    def constantexpr(self, ast):
        return ast

    def builtinexpr(self, ast):
        return ast

    def numericexpr(self, ast):
        return ast

    def characterexpr(self, ast):
        return ast

    def stringexpr(self, ast):
        return ast

    def name(self, ast):
        return ast

    def ALPHA(self, ast):
        return ast

    def NUMERIC(self, ast):
        return ast

    def CHARACTERCONSTCHAR(self, ast):
        return ast

    def STRINGCONSTCHAR(self, ast):
        return ast

    def exprlist(self, ast):
        return ast

    def namelist(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = BParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in BParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for B.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )



def test_auto_vector(check_output):
    check_output('''
        main() {
            extrn putchar;
            auto v[2], i;
            v[0] = 'a';
            v[1] = 's';
            v[2] = 't';
            i = 0;
            while(i <= 2) {
                putchar(v[i]); i =+ 1;
            }
        }
    ''', 'ast')

def test_auto_with_strings(check_output):
    check_output('''
        main() {
            extrn greeting, putstr;
            auto v[2], i;
            v[0] = greeting;
            v[1] = ", ";
            v[2] = "world!*n";
            i = 0;
            while(i <= 2) {
                putstr(v[i]); i =+ 1;
            }
        }
        greeting "hello";
    ''', 'hello, world!\n')

def test_uninit_external_vector(check_output):
    # Technically the initial contents of v are undefined but our implementation
    # always fills external vectors with zeros.
    check_output('''
        main() {
            extrn v, putnumb;
            auto i;
            i = 0;
            v[2] = 8;
            while(i <= 4) { putnumb(v[i]); i =+ 1; }
        }
        v[4];
    ''', '00800')

def test_init_external_vector(check_output):
    check_output('''
        main() {
            extrn v, putnumb;
            auto i;
            i = 0;
            v[2] = 8;
            while(i <= 4) { putnumb(v[i]); i =+ 1; }
        }
        v[4] 12, 34;
    ''', '1234800')

def test_init_external_str_vector(check_output):
    check_output('''
        main() {
            extrn v, putstr;
            auto i;
            i = 0;
            while(i <= 2) { putstr(v[i]); i =+ 1; }
        }
        v[2] "one", "two", "three";
    ''', 'onetwothree')

def test_init_vector_no_count(check_output):
    check_output('''
        main() {
            extrn v, putnumb;
            auto i;
            i = 0;
            while(i <= 2) { putnumb(v[i]); i =+ 1; }
        }
        v[] 1, 2, 3;
    ''', '123')

def test_init_vector_too_few_count(check_output):
    check_output('''
        main() {
            extrn v, putnumb;
            auto i;
            i = 0;
            while(i <= 2) { putnumb(v[i]); i =+ 1; }
        }
        v[1] 1, 2, 3;
    ''', '123')

def test_init_vector_too_many_count(check_output):
    check_output('''
        main() {
            extrn v, putnumb;
            auto i;
            i = 0;
            while(i <= 6) { putnumb(v[i]); i =+ 1; }
        }
        v[6] 1, 2, 3;
    ''', '1230000')



_PROGRAM = '''
    main() {
        extrn putstr, msg;
        auto i;
        i = 0;
        while(i < 3) { putstr(msg[i++]); }
    }
    msg [] "Hello,", " world!", "*n";
'''

def _parse(make_node, source=None, start='program'):
    """Parse B source code using a specified AST node constructor. If source is
    omitted, use _PROGRAM."""
    from rbc.parser import BParser
    from rbc.semantics import BSemantics
    source = source if source is not None else _PROGRAM
    return BParser().parse(source, start, semantics=BSemantics(make_node))

def test_ordered_dict():
    from rbc.dumpast import make_ordered_dict_ast_node
    node = _parse(make_ordered_dict_ast_node)
    assert '_type' in node

def test_json():
    from rbc.dumpast import make_ordered_dict_ast_node, ASTJSONEncoder
    node = _parse(make_ordered_dict_ast_node)
    assert len(ASTJSONEncoder().encode(node)) > 0

def test_graphviz():
    try:
        from io import StringIO
    except ImportError:
        from StringIO import StringIO
    from rbc.dumpast import make_ordered_dict_ast_node, GraphvizAST
    ast = GraphvizAST()
    node = _parse(ast.make_node)
    fobj = StringIO()
    ast.emit(fobj)
    assert len(fobj.getvalue()) > 0




from __future__ import print_function
import contextlib
from future.moves import collections

from llvmlite import ir

# Emitting code from the AST
# ==========================
#
# The LLVM code for the program is emitted after the program has been fully
# parsed. This is required because B functions may refer to functions and
# external variables which have not yet been defined in the program.
#
# LLVM code is emitted within an "emit context". This is some mutable state
# which is used to keep important information on the program and the current
# state of the LLVM code emission.

class EmitContext(object):
    """A context is initialised with an llvm Target and TargetMachine instance.

    """
    def __init__(self, target, machine):
        # Record target and machine
        self.target = target
        self.machine = machine

        # We choose the word type to be an integer with the same size as a
        # pointer to i8. The word size is expressed in bytes
        word_size = ir.IntType(8).as_pointer().get_abi_size(
            self.machine.target_data)

        # Define the word type and bytes-per-word appropriately
        self.word_type = ir.IntType(word_size * 8)
        self.bytes_per_word = word_size

        # Define the type used for constructor function records
        priority_type = ir.IntType(32)
        ctor_func_type = ir.FunctionType(ir.VoidType(), [])
        data_ptr_type = ir.PointerType(ir.IntType(8))
        self.ctor_record_type = ir.LiteralStructType([
            priority_type, ctor_func_type.as_pointer(), data_ptr_type])

        # Initialise context attributes
        self.module = None
        self.global_scope = {}
        self.externals = {}
        self.scope = collections.ChainMap(self.global_scope)
        self.builder = None
        self.string_constants = {}
        self.ctor_records = []

        # Block at end of current switch/which statement
        self.break_block = None

        # LLVM Value of condition in current switch statement
        self.switch_val = None

        # Block which new switch tests should be appended to
        self.switch_block = None

        # Labels are mappings from names to.basic blocks A goto branches to the
        # block. The labels dict is only created within functions.
        self.labels = None

        # Callables which should be called after all code has been emitted
        self.post_emit_hooks = []

        # Flag to indicate when one is within the emitting_code() context.
        self._is_emitting = False

    # There are tasks which must be done to finalise the LLVM module after all
    # code is emitted. Users of the context need not worry about what this
    # finalisation entails but it's important that the finalisation is
    # performed. The emitting_code() method returns a context manager which
    # ensures that the finalisation is performed.

    @contextlib.contextmanager
    def emitting_code(self):
        """A context manager which takes care of any pre- and post-emit tasks.
        (Such as emitting constructor functions.) Only emit code within this
        context.

        """
        assert not self._is_emitting
        assert self.module is None

        if self.module is not None:
            raise RuntimeError('Cannot emit twice using the same EmitContext')

        # Create module for program
        self.module = ir.Module()

        # Set module's triple and data layout from target
        self.module.triple = self.target.triple
        self.module.data_layout = str(self.machine.target_data)

        self._is_emitting = True
        yield
        self._is_emitting = False

        # Create the global variable constructors if necessary
        if len(self.ctor_records) > 0:
            ctor_array_type = ir.ArrayType(
                self.ctor_record_type, len(self.ctor_records))

            var = ir.GlobalVariable(self.module, ctor_array_type,
                                    'llvm.global_ctors')
            var.linkage = 'appending'
            var.initializer = ir.Constant(ctor_array_type, self.ctor_records)

        # Call any post-commit hooks
        for hook in self.post_emit_hooks:
            hook(self)
        self.post_emit_hooks = []

    # Scopes
    # ======
    #
    # Scopes associate names with lvalues. The addresses associated with lvalues
    # never change. "Assigning" to a variable involves writing a new value at
    # the associated address. Variables are simply the lvalues in scopes
    # retrieved by name but the name is only looked up at emit()-time. This is
    # to allow forward-references to in-scope but yet-to-be-declared variables
    # such as global definitions.
    #
    # Scopes are implemented as dict-like objects with string keys and LValue
    # values.

    @contextlib.contextmanager
    def in_child_scope(self):
        """A context manager which switches to a new local scope which is a
        child of the current scope.

        """
        old_scope, self.scope = self.scope, self.scope.new_child()
        yield
        self.scope = old_scope

    # Instruction building
    # ====================
    #
    # LLVM instructions are emitted via a "Instruction Builder". There is only a
    # builder set when there is a block to insert code in. The "builder"
    # attribute should be set to the instruction builder while emitting
    # top-level definitions containing code. If should be None if there is no
    # currently defined point to insert code.

    @contextlib.contextmanager
    def new_function_body(self, entry_block):
        """A context manager which sets values in the context ready for emitting
        a function body. It takes the basic block corresponding to the entry
        point.

        """
        # Create a new scope
        builder = ir.IRBuilder(entry_block)
        old_builder, self.builder = self.builder, builder
        old_labels, self.labels = self.labels, {}
        with self.in_child_scope():
            yield
        self.labels = old_labels
        self.builder = old_builder

    @contextlib.contextmanager
    def setting_break_block(self, block):
        old_block, self.break_block = self.break_block, block
        yield
        self.break_block = old_block

    @contextlib.contextmanager
    def setting_switch_context(self, val, block):
        old_val, self.switch_val = self.switch_val, val
        old_block, self.switch_block = self.switch_block, block
        yield
        self.switch_block = old_block
        self.switch_val = old_val

# Symbol naming
# =============
#
# B allows for global externally visible symbols such as functions. To avoid
# clashes with the C world, we mangle the symbol names. By default the B world
# can't see C and vice versa. This is done by prefixing all B symbols with "b."
# which renders them invalid as C identifiers.

def mangle_symbol_name(name):
    """Mangle an external symbol name for use with B."""
    return 'b.' + name

# Addresses and pointers and words, oh my!
# ========================================
#
# Although B lacks what would today be called a type system (every object is
# of type "word") there is an implicit one in that addresses are assumed to
# be word oriented and thus "address" + "word" should really be "address" +
# (word size * "word") in a byte-oriented architecture. We need to tackle
# this since "a[b]" is syntactic sugar for "*(a + b)" and we don't know at
# emit time which of a and/or b are pointers.  This is further complicated
# by the fact that *neither* of a or b *need* be pointers, "1[2]" is a valid
# vector expression in B, albeit one likely to lead to an invalid memory
# access.
#
# The most straight-forward approach is to have address values be stored
# word-oriented which requires that the alignment of the target be suitable.
# This also necessitates the use of constructor functions and wrappers to
# shuffle between "addresses" and pointers used by LLVM.

def address_to_llvm_ptr(context, address_val, ptr_type):
    """Cast a llvm Value representing a word into a pointer. Performs the
    appropriate conversion from word-oriented addresses to llvm's
    byte-oriented addresses. Requires builder to be not None.

    This method will arrange that

        address_to_llvm_ptr(llvm_ptr_to_address(ptr_val), ptr_type)

    is elided into a simple bitcast. This implies that ptr_val must be word
    aligned. It's the responsibility of the code generator to make sure
    that's the case.

    """
    # If this address came from a pointer, just return the punned pointer.
    if hasattr(address_val, 'b_ptr'):
        return context.builder.bitcast(address_val.b_ptr, ptr_type)

    bpw_const = ir.Constant(context.word_type, context.bytes_per_word)
    byte_address = context.builder.mul(address_val, bpw_const,
                                       flags=['nuw', 'nsw'])
    ptr_val = context.builder.inttoptr(byte_address, ptr_type)

    # HACK: we tag the value with a 'b_address' attribute so that
    # llvm_ptr_to_address() can elide address to pointer followed by pointer
    # to address conversion.
    ptr_val.b_address = address_val

    return ptr_val

def llvm_ptr_to_address(context, ptr_val):
    """Cast a llvm Value representing a byte-oriented pointer to a word
    representing a word-oriented address. Requires builder to be not None.

    This method will arrange that

        llvm_ptr_to_address(address_to_llvm_ptr(address_val, ptr_type))

    is elided to address_val irrespective of ptr_type.

    """
    # If this pointer came from an address, just return the original address
    # value.
    if hasattr(ptr_val, 'b_address'):
        return ptr_val.b_address

    bpw_const = ir.Constant(context.word_type, context.bytes_per_word)
    byte_address = context.builder.ptrtoint(ptr_val, context.word_type)
    address_val = context.builder.udiv(byte_address, bpw_const,
                                       flags=['exact'])

    # HACK: we tag the value with a 'b_ptr' attribute so that
    # address_to_llvm_ptr() can elide pointer to address followed by address
    # to pointer conversion.
    address_val.b_ptr = ptr_val

    return address_val

# Convenience functions
# =====================
#
# Code generation can be quite repetitive. Define some functions which express
# common functionality.

def if_else(context, cond):
    """Convenience wrapper around llvm.ir.Builder.if_else. Takes a single AST
    node representing an expression which should be treated as a condition and
    evaluates that node.

    """
    cond_val = cond.emit(context)
    zero = ir.Constant(context.word_type, 0)
    bool_val = context.builder.icmp_signed('!=', cond_val, zero)
    return context.builder.if_else(bool_val)

def create_constructor(context, priority=0, name_hint=None):
    """Create a function which is called on module load. Functions are called in
    increasing order or priority. (The order of functions with equal priority is
    undefined.)

    """
    ctor_record_type = context.ctor_record_type
    priority_type, ctor_func_ptr_type, data_ptr_type = ctor_record_type.elements

    # Create the constructor function
    if name_hint is None:
        func_name = context.module.get_unique_name('__ctor')
    else:
        func_name = context.module.get_unique_name(
            '__ctor.{}'.format(name_hint))
    func = ir.Function(context.module, ctor_func_ptr_type.pointee, func_name)
    func.linkage = 'private'

    context.ctor_records.append(ir.Constant(ctor_record_type, [
        ir.Constant(priority_type, priority), func,
        ir.Constant(data_ptr_type, None)]))

    return func

def create_aligned_global(module, type_, name):
    """Construct and return an ir.GlobalVariable instance which is guaranteed to
    have a word-aligned pointer. The symbol name is automatically mangled to
    avoid collision with C symbols.

    """
    return _ModifiedGlobalVariable(module, type_, mangle_symbol_name(name))

class _ModifiedGlobalVariable(ir.GlobalVariable):
    """A derived class based on llvm.ir.GlobalVariable with an added attribute,
    "modifiers" which is a sequence of string constants appended to the variable
    declaration separated by commas. If modifiers is of non-zero length, it also
    has a leading comma. This is useful for specifying alignment, etc.

    Note that this class makes use of knowledge of the llvm.ir internals and is
    not a good long-term solution.

    """
    def __init__(self, *args, **kwargs):
        ir.GlobalVariable.__init__(self, *args, **kwargs)
        self.modifiers = []

    def descr(self, buf):
        ir.GlobalVariable.descr(self, buf)
        if len(self.modifiers) > 0:
            print(',', file=buf)
            print(','.join(self.modifiers), file=buf)




def test_empty_module(compile_b):
    compile_b('')

def test_empty_function_is_defined(compile_b):
    compile_b('f(){}')

def test_return_argument(compile_b):
    compile_b('f(arg){ return (arg); }')

def test_return_constant(compile_b):
    compile_b('f(){ return (42); }')

def test_return_nothing(compile_b):
    compile_b('f(){ return; }')

def test_return_function(compile_b):
    compile_b('f1(){} f2(){ return (f1); }')

def test_return_not_yet_defined_function(compile_b):
    compile_b('f1(){ return(f2); } f2(){}')

def test_return_simple_extern(compile_b):
    compile_b('f(){ extrn A; return(A); } A;')

def test_return_deref_extern(compile_b):
    compile_b('f1(){} f2(){return(&f1);}')

def test_extern_function_call(compile_b):
    compile_b('f1(){ extrn putchar; putchar(65); }')

def test_names_with_backspace(compile_b):
    compile_b('m\byfunc(){ }')

def test_simple_expression(check_output):
    check_output('main() { extrn putnumb; putnumb(2-3-4); }', '-5')
    check_output('main() { extrn putnumb; putnumb(4+10*2); }', '24')
    check_output('main() { extrn putnumb; putnumb(5 & 2); }', '0')
    check_output('main() { extrn putnumb; putnumb(7 & 2); }', '2')

def test_func_call(check_output):
    check_output('''
        main() { extrn putnumb; putnumb(sub(3,4)); }
        sub(a, b) { return (a-b); }
    ''', '-1')

def test_simple_auto(check_output):
    check_output('''
        main() {
            extrn putnumb;
            auto a, b;
            a = 4; b = 10;
            putnumb(b-a);
        }
    ''', '6')

def test_hello_world(check_output):
    check_output('''
        main() {
            extrn putchar;
            putchar('hell');
            putchar('o, w');
            putchar('orld');
        }
    ''', 'hello, world')

def test_recursion(check_output):
    check_output('''
        fact(n) if(n == 0) return(1); else return (n*fact(n-1));
        check(a, b) {
            extrn putchar;
            if(a != b)
                putchar('x');
            else
                putchar('.');
        }
        main() {
            check(fact(0), 1);
            check(fact(1), 1);
            check(fact(2), 2);
            check(fact(3), 6);
            check(fact(4), 24);
        }
    ''', '.....')

def test_zero_extrn(check_output):
    check_output('''
        main() {
            extrn a, putnumb;
            putnumb(a + 5);
        }
        a;
    ''', '5')

def test_extrn_hello_world(check_output):
    check_output('''
        main() {
            extrn a, b, c, putchar;
            putchar(a); putchar(b); putchar(c); putchar('!*n');
        }
        a 'hell';
        b 'o, w';
        c 'orld';
    ''', 'hello, world!\n')

def test_extrn_hello_world_2(check_output):
    check_output('''
        main( ) {
            extrn a,b,c,d;
            put2char(a,b) ;
            put2char(c,d) ;
        }
        put2char(x,y) {
            extrn putchar;
            putchar(x);
            putchar(y);
        }
        a 'hell'; b 'o, w'; c 'orld'; d '!*n';
    ''', 'hello, world!\n')

def test_setting_extrn(check_output):
    check_output('''
        main() {
            extrn putnumb, val;
            addval(5);
            addval(3);
            putnumb(val);
        }
        addval(x) {
            extrn val;
            val =+ x;
        }
        val;
    ''', '8')




from __future__ import print_function

from llvmlite import ir

import rbc.exception as exc

from .astnode import ast_node, needs_builder, ASTNode
from .context import create_aligned_global, if_else
from .expression import LLVMPointerValue

def get_or_create_global(context, name):
    """Retrieve the LValue and llvm GlobalValue associated with an external
    variable with the specified name. If there is no defined external variable,
    a new uninitialised global variable is created which must be resolved at
    link time.

    """
    # Return the existing extern if present.
    if name in context.externals:
        return context.externals[name]

    # Create a new variable in the module and add it to the externals.
    value = create_aligned_global(context.module, context.word_type, name)
    value.modifiers = ['align {}'.format(context.bytes_per_word)]

    lvalue = LLVMPointerValue(value=value).dereference()
    context.externals[name] = lvalue
    return lvalue

# Statements
# ==========
#
# Statements, unlike rvalues or lvalues, do not return llvm Values when emitted.

@ast_node
class AutoStatement(ASTNode):
    """An auto variable is automatically allocated onto the stack."""
    @needs_builder
    def emit(self, context):
        val = context.builder.alloca(context.word_type, name=self.name)
        context.scope[self.name] = LLVMPointerValue(value=val).dereference()

@ast_node
class AutoVectorStatement(ASTNode):
    """An auto variable is automatically allocated onto the stack."""
    @needs_builder
    def emit(self, context):
        # Curiously, B uses the "maximum index" when declaring vectors. This
        # mean, for example, that "auto v[3]" should allocate storage for *4*
        # words on the stack rather than 3 as would seem more likely.
        vector_length = 1 + self.maxidx.value

        # Allocate values and record in scope
        val = context.builder.alloca(context.word_type, size=vector_length,
                                     name=self.name)

        # In contrast to non-vector auto variables, the "value" of a vecotr auto
        # is the actual underlying pointer rather than the dereferenced pointer.
        context.scope[self.name] = LLVMPointerValue(value=val)

@ast_node
class ExtrnStatement(ASTNode):
    def emit(self, context):
        context.scope[self.name] = get_or_create_global(context, self.name)

@ast_node
class MultipartStatement(ASTNode):
    """A statement which is like a CompoundStatement but there is no change of
    scope."""
    def emit(self, context):
        for statement in self.statements:
            statement.emit(context)

@ast_node
class CompoundStatement(MultipartStatement):
    """A group of statements emitted within a brand new scope. This isn't in any
    of the reference manuals I can find but it is C-like which suggests it may
    have been B-like.

    """
    def emit(self, context):
        with context.in_child_scope():
            MultipartStatement.emit(self, context)

@ast_node
class ReturnStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        if self.return_value is None:
            # The default return value is 0
            ret_val = ir.Constant(context.word_type, 0)
        else:
            ret_val = self.return_value.emit(context)

        # Emit the return instruction
        context.builder.ret(ret_val)

        # A return instruction is a terminator and so we should create a new
        # block which any post-return instructions are appended to.
        new_block = context.builder.append_basic_block('post_return')
        context.builder.position_at_end(new_block)

@ast_node
class WhileStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        # Create basic blocks for builder
        while_cond = context.builder.append_basic_block('while')
        while_body = context.builder.append_basic_block('whilethen')
        while_end = context.builder.append_basic_block('whileend')

        # Initially, branch into while loop condition test
        context.builder.branch(while_cond)

        # Position at start of condition block and emit condition and
        # test
        context.builder.position_at_end(while_cond)
        cond = self.cond.emit(context)
        zero = ir.Constant(context.word_type, 0)
        cond_is_not_zero = context.builder.icmp_signed('!=', cond, zero)

        # Conditionally branch to body or end of while
        context.builder.cbranch(cond_is_not_zero, while_body, while_end)

        # Position in body and emit
        context.builder.position_at_end(while_body)
        with context.setting_break_block(while_end):
            self.body.emit(context)

        # Branch back to condition
        context.builder.branch(while_cond)

        # Position after loop for further instructions
        context.builder.position_at_end(while_end)

@ast_node
class IfStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        with if_else(context, self.cond) as (then, otherwise):
            with then:
                self.then.emit(context)
            with otherwise:
                if self.otherwise is not None:
                    self.otherwise.emit(context)

@ast_node
class ExpressionStatement(ASTNode):
    def emit(self, context):
        """An expression statement simply evaluates its expression and discards
        the result."""
        self.expression.emit(context)

@ast_node
class NullStatement(ASTNode):
    def emit(self, context):
        """A null statement does nothing."""

@ast_node
class LabelStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        # Create a new basic block and record
        assert self.label not in context.labels
        new_block = context.builder.append_basic_block(self.label)
        context.labels[self.label] = new_block

        # Unconditionally branch to new block
        context.builder.branch(new_block)

        # Move context to new block
        context.builder.position_at_end(new_block)

        # Emit the following statement
        self.statement.emit(context)

@ast_node
class GotoStatement(ASTNode):
    """The goto statement becomes an unconditional branch. Since labels may be
    defined after the goto statement is emitted, the goto statement creates a
    new basic block for the branch to be appended to but does not immediately
    emit the branch instruction. Instead, a post-emit hook is added to emit the
    branch instruction once all labels have been defined.

    """
    def emit(self, context):
        current_block = context.builder.block
        current_builder = context.builder
        labels = context.labels

        # We create a post-emit hook so that the goto's label is resolved after
        # all the rest of the code is emitted.
        def hook(_):
            # Move back to the correct block
            current_builder.position_at_end(current_block)

            # Reset scope and retrieve block
            try:
                block = labels[self.label]
            except KeyError:
                raise exc.SemanticError('No such label: {}'.format(self.label))

            # Branch to label
            current_builder.branch(block)
        context.post_emit_hooks.append(hook)

        # Create a new block for appending future instructions
        new_block = context.builder.append_basic_block('post_goto')
        context.builder.position_at_end(new_block)

@ast_node
class BreakStatement(ASTNode):
    """The break statement is very much like a goto statement in that it turns
    into a conditional branch. Unlike a goto, the block to branch to is
    specified by the compound statement node when it is emitted. A break
    instruction branches to the basic block immediately after the current
    compound statement. If there is no compound statement being emitted, break
    is a no-op.

    """
    def emit(self, context):
        # Break statements are no-ops if we've no destination
        if context.break_block is None:
            return

        # Jump straight to the end of context
        context.builder.branch(context.break_block)

        # Add a new block for any following instructions
        new_block = context.builder.append_basic_block('post_break')
        context.builder.position_at_end(new_block)

@ast_node
class SwitchStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        # Evaluate the switches condition value
        switch_val = self.rvalue.emit(context)

        # Add a block used to test condition
        test_block = context.builder.append_basic_block('switch_test')
        if not context.builder.block.is_terminated:
            context.builder.branch(test_block)

        # Create a new block for the switch
        switch_entry_block = context.builder.append_basic_block('switch_entry')
        end_block = context.builder.append_basic_block('switch_end')
        context.builder.position_at_end(switch_entry_block)

        # Emit the body with the switch context set
        with context.setting_switch_context(switch_val, test_block):
            with context.setting_break_block(end_block):
                self.body.emit(context)

            # Fall through to end block
            if not context.builder.block.is_terminated:
                context.builder.branch(end_block)

            # At the end of the test block, branch to end
            context.builder.position_at_end(context.switch_block)
            if not context.builder.block.is_terminated:
                context.builder.branch(end_block)

        # Set end block as new builder position
        context.builder.position_at_end(end_block)

@ast_node
class CaseStatement(ASTNode):
    @needs_builder
    def emit(self, context):
        if context.switch_val is None:
            raise exc.SemanticError('case outside of switch')

        # Create a new block for case
        case_block = context.builder.append_basic_block('case')
        if not context.builder.block.is_terminated:
            context.builder.branch(case_block)

        # Now move to test block
        context.builder.position_at_end(context.switch_block)

        if self.cond is None:
            # If there's no condition, then this is a default block
            context.builder.branch(case_block)
        elif not context.builder.block.is_terminated:
            # Test condition
            case_val = self.cond.emit(context)
            is_eq = context.builder.icmp_signed(
                '==', context.switch_val, case_val)
            next_block = context.builder.append_basic_block('case.else')
            context.builder.cbranch(is_eq, case_block, next_block)
            context.builder.position_at_end(next_block)
            context.switch_block = next_block

        # Emit next statements at end of case block
        context.builder.position_at_end(case_block)
        self.then.emit(context)




def test_simple_comment(check_output):
    check_output('''
        main() {
            extrn putchar;
            /* a comment */
            putchar('a');
        }
    ''', 'a')

def test_comment_stops_at_first_terminator(check_output):
    check_output('''
        main() {
            extrn putchar;
            /* a comment */
            putchar('a');
            /* another comment */
        }
    ''', 'a')


def test_comment_accepts_initial_asterisk(check_output):
    check_output('''
        main() {
            extrn putchar;
            /** a comment */
            putchar('a');
        }
    ''', 'a')


def test_comment_accepts_final_asterisk(check_output):
    check_output('''
        main() {
            extrn putchar;
            /* a comment **/
            putchar('a');
        }
    ''', 'a')


def test_comment_accepts_medial_asterisk(check_output):
    check_output('''
        main() {
            extrn putchar;
            /* a * comment */
            putchar('a');
        }
    ''', 'a')

def test_comment_accepts_newline(check_output):
    check_output('''
        main() {
            extrn putchar;
            /* a
            multi
            line
            comment */
            putchar('a');
        }
    ''', 'a')




def test_constant_decimal(check_expr):
    check_expr('25', 25)

def test_constant_octal(check_expr):
    # Real programmers think Halloween is on the same day as Christmas
    check_expr('031', 25)

def test_add_mult_precedence(check_expr):
    # Subtraction binds LTR
    check_expr('2-3-4', -5)
    # Division binds LTR
    check_expr('20/5/2', 2)
    # Addition higher precedence than multiplication
    check_expr('2*4+5*3', 23)
    # Brackets override precedence
    check_expr('2*(4+5)*3', 54)

def test_basic_relational(check_expr):
    check_expr('6>5', '1')
    check_expr('5>5', '0')
    check_expr('4>5', '0')
    check_expr('4>=5', '0')
    check_expr('5>=5', '1')
    check_expr('6>=5', '1')
    check_expr('6<5', '0')
    check_expr('5<5', '0')
    check_expr('4<5', '1')
    check_expr('4<=5', '1')
    check_expr('5<=5', '1')
    check_expr('6<=5', '0')
    check_expr('6==5', '0')
    check_expr('5==5', '1')
    check_expr('6!=5', '1')
    check_expr('5!=5', '0')

def test_char_constant(check_expr):
    check_expr("''", '0')
    check_expr("'A'", '65')
    check_expr("'ab'", '24930')
    check_expr("'a*0'", '24832')
    check_expr("'*''", '39')
    check_expr("' x '", '2127904')

def test_unary_binding(check_expr):
    check_expr('-~0', '1')

def test_assignment_op(check_output):
    def check(op, expected):
        check_output('main(){extrn putnumb;auto a;a=10;a'+op+' 2;putnumb(a);}',
                     expected)
    check('=', '2')
    check('=-', '8')
    check('=+', '12')
    check('=*', '20')
    check('=/', '5')
    check('=%', '0')

def test_assignment_op_has_value(check_output):
    check_output('''
        main() {
            extrn a, putnumb;
            auto b, c;
            putnumb(c=b=(a==23));
            putnumb(c);
            putnumb(b);
            putnumb(a);
        }
        a 23;
    ''', '11123')

def test_builtin_bytes_per_word(output_from):
    bpw = output_from('''
        main() {
            extrn putnumb;
            putnumb(__bytes_per_word);
        }
    ''')
    assert bpw != ''
    bpw_val = int(bpw)
    assert bpw_val > 0

def test_inline_string_constants(check_output):
    check_output('''
        main() {
            extrn putstr;
            putstr("hello, "); putstr("world!"); putstr("*n");
        }
    ''', 'hello, world!\n')

def test_auto_string_constants(check_output):
    check_output('''
        main() {
            extrn putstr;
            auto s1, s2;
            s1 = "hello, "; s2 = "world!*n";
            putstr(s1); putstr(s2);
        }
    ''', 'hello, world!\n')

def test_extern_string_constants(check_output):
    check_output('''
        main() {
            extrn putstr, s1, s2;
            putstr(s1); putstr(s2);
        }
        s1 "hello, "; s2 "world!*n";
    ''', 'hello, world!\n')

def test_conditional_op(check_expr):
    check_expr('5>6?2:3', '3')
    check_expr('5<6?2:3', '2')
    check_expr('0?2:3?4:5', '4')

def test_conditional_evaluates_one_side(check_output):
    check_output('''
        main() {
            1<2 ? f(10):f(11);
            5>9 ? f(20):f(21);
        }
        f(x) {
            extrn putnumb;
            putnumb(x);
            return (x);
        }
    ''', '1021')




from __future__ import print_function

from llvmlite import ir

from .astnode import ast_node, ASTNode

from .context import (
    address_to_llvm_ptr, create_constructor, create_aligned_global,
    mangle_symbol_name
)

from .expression import ConstantIntValue, LLVMPointerValue

# Globals
# =======
#
# Globals are emittables which have an additional "declare" method. All globals
# have their declare() method called before any emittables have emit() called.
# They should use this opportunity to register themselves with the current
# context.
#
# A global should expect the context's builder attribute to be None.

@ast_node
class SimpleDefinition(ASTNode):
    """An initialised external variable."""
    def __init__(self, **kwargs):
        ASTNode.__init__(self, **kwargs)
        self._lvalue = None

    def declare(self, context):
        # A simple global definition is represented in the llvm IR as a pointer
        # to the global value. Thus we may use a LLVMPointerValue to represent
        # the address of the lvalue.
        value = create_aligned_global(
            context.module, context.word_type, self.name)
        value.modifiers = ['align {}'.format(context.bytes_per_word)]

        # Set variable's initialiser. Don't bother with a constructor function
        # if the initialiser is a constant integer
        if self.init is not None and isinstance(self.init, ConstantIntValue):
            init_val = self.init.value
        else:
            init_val = 0
        value.initializer = ir.Constant(context.word_type, init_val)

        # Register this variable as an external symbol.
        self._lvalue = LLVMPointerValue(value=value).dereference()
        context.externals[self.name] = self._lvalue

    def emit(self, context):
        assert self._lvalue is not None

        # If we have no initialiser, we do not need to emit a constructor
        if self.init is None or isinstance(self.init, ConstantIntValue):
            return

        # Initialisers may themselves be global variables. In which case we need
        # to make sure we initialise them in the correct order.
        func = create_constructor(context, priority=0, name_hint=self.name)

        # Create entry block for function and associated builder
        block = func.append_basic_block(name='entry')

        # Assign the variable's value
        with context.new_function_body(block):
            init = self.init.emit(context)
            lvalue_address = self._lvalue.reference().emit(context)
            value_ptr = address_to_llvm_ptr(
                context, lvalue_address, context.word_type.as_pointer())
            context.builder.store(init, value_ptr)
            context.builder.ret_void()

@ast_node
class VectorDefinition(ASTNode):
    def __init__(self, **kwargs):
        ASTNode.__init__(self, **kwargs)
        self._lvalue = None

    def declare(self, context):
        # SCJ: "The actual size of the vector is the maximum of constant+1 and
        # the number of initial values. Any vector elements which are not
        # explicitly initialized have undefined values."
        if self.maxidx is not None:
            maxidx = self.maxidx.value
        else:
            maxidx = 0
        n_elems = max(len(self.ivals), maxidx + 1)

        # Initialise the value with zeros
        value_type = ir.ArrayType(context.word_type, n_elems)
        value = create_aligned_global(context.module, value_type, self.name)
        value.modifiers = ['align {}'.format(context.bytes_per_word)]
        value.initializer = ir.Constant(value_type, None)

        # Register this variable as an external symbol. Note that the pointer
        # value itself is registered unlike SimpleDefinition.
        self._lvalue = LLVMPointerValue(value=value)
        context.externals[self.name] = self._lvalue

    def emit(self, context):
        assert self._lvalue is not None
        if len(self.ivals) == 0:
            # No initialisation required
            return

        # Initialisers may themselves be global variables. In which case we need
        # to make sure we initialise them in the correct order.
        func = create_constructor(context, priority=0, name_hint=self.name)

        # Create entry block for function and associated builder
        block = func.append_basic_block(name='entry')

        # Assign the variable's values
        with context.new_function_body(block):
            lvalue = self._lvalue.emit(context)
            value_ptr = address_to_llvm_ptr(
                context, lvalue, context.word_type.as_pointer())
            for idx, val in enumerate(self.ivals):
                llvm_val = val.emit(context)
                idx_val = ir.Constant(context.word_type, idx)
                dest_ptr = context.builder.gep(value_ptr, [idx_val])
                context.builder.store(llvm_val, dest_ptr)
            context.builder.ret_void()

@ast_node
class FunctionDefinition(ASTNode):
    def __init__(self, **kwargs):
        ASTNode.__init__(self, **kwargs)
        self._func = None

    def declare(self, context):
        # Check we've not been declared already
        assert self._func is None

        # Create a new function type for this function
        word_type = context.word_type
        n_args = len(self.arg_names)
        func_type = ir.FunctionType(word_type, (word_type,) * n_args)

        # Create the function in the module and add to the global scope
        symbol_name = mangle_symbol_name(self.name)
        self._func = ir.Function(context.module, func_type, name=symbol_name)
        context.global_scope[self.name] = LLVMPointerValue(
            value=self._func).dereference()

    def emit(self, context):
        # Create entry block for function and associated builder
        block = self._func.append_basic_block(name='entry')
        with context.new_function_body(block):
            # Add function arguments to the function scope
            for arg_name, arg_value in zip(self.arg_names, self._func.args):
                arg_value.name = arg_name

                # Allocate stack variable for this argument and copy argument
                stack_var = context.builder.alloca(
                    context.word_type, name=arg_name)
                context.builder.store(arg_value, stack_var)

                # Store the stack variable in the scope
                context.scope[arg_name] = LLVMPointerValue(
                    value=stack_var).dereference()

            # Emit the function body
            self.body.emit(context)

            # All functions implicitly return 0 if there's no other return
            if not context.builder.block.is_terminated:
                context.builder.ret(ir.Constant(context.word_type, 0))



def test_basic_switch(check_output):
    check_output('''
        main() {
            extrn putchar;;
            auto i;
            i = 0; while(i <= 4) {
                describe(i);
                putchar('*n');
                ++i;
            }
        }
        describe(val) {
            extrn putstr, putnumb;
            putnumb(val); putstr(" is ");
            switch(val) {
                case 0: putstr("zero"); break;
                case 1: putstr("one"); break;
                default: putstr("many"); break;
            }
        }
    ''', '0 is zero\n1 is one\n2 is many\n3 is many\n4 is many\n')

def test_fallthrough(check_output):
    check_output('''
        main() {
            extrn putchar;;
            auto i;
            i = 0; while(i <= 4) {
                describe(i);
                putchar('*n');
                ++i;
            }
        }
        describe(val) {
            extrn putstr, putnumb;
            putnumb(val); putstr(" is ");
            switch(val) {
                case 0: putstr("zero");
                case 1: putstr("one"); break;
                default: putstr("many");
            }
        }
    ''', '0 is zeroone\n1 is one\n2 is many\n3 is many\n4 is many\n')




"""
Convert parsed grammar into AST nodes.

"""
from future.builtins import bytes

# Semantics
# =========
#
# A semantics object is used by the parser to construct AST nodes. Our semantics
# converts parser AST nodes which reflect the mechanism of the parser into AST
# nodes which better reflect the semantics of the program. For example, in the
# parser binary operator expressions are parsed as a higher precedence
# expression followed by zero or more operator, expression pairs. The semantics
# object converts those into zero or more BinaryOpValue AST nodes.

class BSemantics(object):
    """The BSemantics class represents the transformation of a parsed B program
    into AST nodes.

    The semantics are provided with a callable which is used to generate new AST
    nodes. The callable should take a single positional argument which is a
    string specifying the AST node name and a set of zero or more keyword
    arguments giving the parameters to the node.

    Args:
        make_node (callable): callable used to make new AST nodes

    """
    def __init__(self, make_node):
        # A callable which takes an AST node name and set of keyword arguments
        # and returns the corresponding AST node object.
        self._node = make_node

    # Programs
    # ========
    #
    # The "program" rule returns the B Program itself.

    def program(self, definitions):
        return self._node('Program', definitions=definitions)

    # Global definitions
    # ==================
    #
    # Programs consist of zero or more top-level definitions. These definitions
    # may be functions or initialised external variables.

    def simpledef(self, ast):
        return self._node('SimpleDefinition', name=ast.name, init=ast.init)

    def vectordef(self, ast):
        ivals = ast.ivals if ast.ivals is not None else []
        return self._node('VectorDefinition', name=ast.name,
                          maxidx=ast.maxidx, ivals=ivals)

    def functiondef(self, ast):
        args = ast.args if ast.args is not None else []
        return self._node(
            'FunctionDefinition', name=ast.name, arg_names=args,
            body=ast.body)

    # Statements
    # ==========
    #
    # A function body is a single statement. However statements may include
    # other statements.

    def _coalesce_statements(self, statements):
        """Accept a sequence of zero or more statements and return an AST node
        which is one of a statement, null statement or multipart statement as
        appropriate.

        """
        if len(statements) == 0:
            return self._node('ExpressionStatement', expression=None)
        elif len(statements) == 1:
            return statements[0]
        else:
            return self._node('MultipartStatement', statements=statements)

    def autostatement(self, substatements):
        return self._coalesce_statements(substatements)

    def autovar(self, ast):
        if ast.maxidx is None:
            return self._node('AutoStatement', name=ast.name)
        return self._node('AutoVectorStatement', name=ast.name,
                          maxidx=ast.maxidx)

    def extrnstatement(self, names):
        statements = [self._node('ExtrnStatement', name=name) for name in names]
        return self._coalesce_statements(statements)

    def compoundstatement(self, statements):
        return self._node('CompoundStatement', statements=statements)

    def ifstatement(self, ast):
        return self._node('IfStatement', cond=ast.cond, then=ast.then,
                          otherwise=ast.otherwise)

    def whilestatement(self, ast):
        return self._node('WhileStatement', cond=ast.cond, body=ast.body)

    def returnstatement(self, ast):
        return self._node('ReturnStatement', return_value=ast.return_value)

    def exprstatement(self, expression):
        return self._node('ExpressionStatement', expression=expression)

    def nullstatement(self, _):
        return self._node('NullStatement')

    def labelstatement(self, ast):
        return self._node('LabelStatement', label=ast.label,
                          statement=ast.statement)

    def gotostatement(self, ast):
        return self._node('GotoStatement', label=ast.label)

    def switchstatement(self, ast):
        return self._node('SwitchStatement', rvalue=ast.rvalue, body=ast.body)

    def casestatement(self, ast):
        return self._node('CaseStatement', cond=ast.cond, then=ast.then)

    def breakstatement(self, _):
        return self._node('BreakStatement')

    # Expressions
    # ===========

    def assignexpr(self, ast):
        if ast.op is None:
            return ast.lhs
        return self._node('AssignmentOpValue', lhs=ast.lhs, op=ast.op,
                          rhs=ast.rhs)

    def condexpr(self, ast):
        if ast.then is None:
            return ast.cond
        return self._node('ConditionalOpValue', cond=ast.cond, then=ast.then,
                          otherwise=ast.otherwise)

    def _leftbinopexpr(self, ast):
        """All left-to-right binary operators are handled similarly."""
        lhs = ast.lhs
        for tail_elem in ast.tail:
            lhs = self._node('BinaryOpValue', lhs=lhs, op=tail_elem.op,
                             rhs=tail_elem.rhs)
        return lhs

    def orexpr(self, ast):
        return self._leftbinopexpr(ast)

    def xorexpr(self, ast):
        return self._leftbinopexpr(ast)

    def andexpr(self, ast):
        return self._leftbinopexpr(ast)

    def eqexpr(self, ast):
        return self._leftbinopexpr(ast)

    def relexpr(self, ast):
        return self._leftbinopexpr(ast)

    def shiftexpr(self, ast):
        return self._leftbinopexpr(ast)

    def addexpr(self, ast):
        return self._leftbinopexpr(ast)

    def multexpr(self, ast):
        return self._leftbinopexpr(ast)

    def unaryexpr(self, ast):
        val = ast.rhs

        # Binding of unary ops is right-to-left
        for op in reversed(ast.rightops):
            val = self._node('RightUnaryOpValue', op=op, lhs=val)

        for op in reversed(ast.leftops):
            val = self._node('LeftUnaryOpValue', op=op, rhs=val)

        return val

    def primaryexpr(self, ast):
        val = ast.head

        if ast.tail is None:
            return val

        # Primary expressions bind left-to-right
        for tail_elem in ast.tail:
            if tail_elem.index is not None:
                # A vector expression is syntactic sugar for addition and
                # dereference.
                addr = self._node('BinaryOpValue', op='+', lhs=val,
                                  rhs=tail_elem.index)
                val = self._node('DereferencedRValue', rvalue=addr)
            else:
                # Otherwise, this is a function call
                args = tail_elem.args if tail_elem.args is not None else []
                val = self._node('FunctionCallValue', func=val, args=args)

        return val

    def variableexpr(self, name):
        return self._node('ScopeValue', name=name)

    def builtinexpr(self, ast):
        return self._node('BuiltinValue', name=ast)

    def numericexpr(self, ast):
        string_value = ''.join(ast)

        # Leading 0 implies octal
        if string_value.startswith('0'):
            int_value = int(string_value, 8)
        else:
            int_value = int(string_value, 10)

        return self._node('ConstantIntValue', value=int_value)

    def characterexpr(self, characters):
        val = 0
        for ch in _expand_escapes(characters):
            val = 0x100 * val + ch
        return self._node('ConstantIntValue', value=val)

    def stringexpr(self, characters):
        str_val = bytes(list(_expand_escapes(characters)))
        return self._node('StringConstantValue', value=str_val)

    # Names
    # =====

    def name(self, ast):
        return ast.head + ''.join(ast.tail)

def _expand_escapes(characters):
    """Return a generator which yields byte-sized character
    values from the sequence *characters*, replacing escape
    sequences.

    """
    ESCAPE_SEQUENCES = {
        '*0': '\0', '*e': '\x04', '*(': '{', '*)': '}',
        '*t': '\t', '**': '*', "*'": "'", '*"': '"',
        '*n': '\n'
    }

    for ch in characters:
        if len(ch) == 1:
            yield ord(ch)
        else:
            try:
                yield ord(ESCAPE_SEQUENCES[ch])
            except KeyError:
                raise ValueError(
                    'Unknown escape sequence: {}'.format(ch))



"""
Usage:
    rbc (-h | --help)
    rbc [-c | -s] [-o FILE] [-O LEVEL] [--emit-llvm] <file>...

Options:
    -h, --help      Show a brief usage summary.
    -o=FILE         Write output to FILE. The default is to use the basename of
                    <file> with an appropriate extension appended.
    -O=LEVEL        Set optimisation level from 0 to 3. [default: 1]
    -c              Generate object file output.
    -s              Generate assembly output.

Advanced options:
    --emit-llvm     Emit LLVM bytecode/assembly rather than native code when -c
                    or -s is specified.

"""
import enum
import os
import sys

import docopt

import rbc.compiler

class OptionError(RuntimeError):
    pass

OutputType = enum.Enum('OutputType', 'object asm executable')

class Options(object):
    def __init__(self, opts):
        self.input_files = opts['<file>']
        self.output_file = opts['-o']

        if opts['-c']:
            self.output_type = OutputType.object
        elif opts['-s']:
            self.output_type = OutputType.asm
        else:
            self.output_type = OutputType.executable

        if self.output_type != OutputType.executable and \
                len(self.input_files) > 1:
            raise OptionError('Only one file with -c or -s option')

        self.emit_llvm = opts['--emit-llvm']

        self.opt_level = int(opts['-O'])
        if self.opt_level < 0 or self.opt_level > 3:
            raise OptionError('Optimisation level must be between 0 and 3.')

def compile_object(output_file, source_file, compiler_options, emit_llvm):
    with open(source_file) as fobj:
        source = fobj.read()
    module_asm = rbc.compiler.compile_b_source(source, compiler_options)
    module = rbc.compiler.optimize_module(module_asm, compiler_options)
    module.name = os.path.basename(source_file)
    with open(output_file, 'wb') as fobj:
        if emit_llvm:
            fobj.write(module.as_bitcode())
        else:
            fobj.write(compiler_options.machine.emit_object(module))

def compile_asm(output_file, source_file, compiler_options, emit_llvm):
    with open(source_file) as fobj:
        source = fobj.read()
    module_asm = rbc.compiler.compile_b_source(source, compiler_options)
    module = rbc.compiler.optimize_module(module_asm, compiler_options)
    module.name = os.path.basename(source_file)
    with open(output_file, 'w') as fobj:
        if emit_llvm:
            fobj.write(str(module))
        else:
            fobj.write(compiler_options.machine.emit_assembly(module))

def main():
    """Main entry point for rbc tool."""
    # Parse CLI opts
    opts = Options(docopt.docopt(__doc__))

    compiler_options = rbc.compiler.CompilerOptions()
    compiler_options.opt_level = opts.opt_level

    if opts.output_type == OutputType.executable:
        if opts.output_file is None:
            output_file = 'a.out'
        else:
            output_file = opts.output_file
        rbc.compiler.compile_and_link(
            output_file, opts.input_files, options=compiler_options)
    elif opts.output_type == OutputType.object:
        output_file = opts.output_file
        if output_file is None:
            out_ext = '.bc' if opts.emit_llvm else '.o'
            output_file = os.path.splitext(opts.input_files[0])[0] + out_ext
        compile_object(output_file, opts.input_files[0], compiler_options,
                       emit_llvm=opts.emit_llvm)
    elif opts.output_type == OutputType.asm:
        output_file = opts.output_file
        if output_file is None:
            out_ext = '.ll' if opts.emit_llvm else '.s'
            output_file = os.path.splitext(opts.input_files[0])[0] + out_ext
        compile_asm(output_file, opts.input_files[0], compiler_options,
                    emit_llvm=opts.emit_llvm)
    else:
        raise RuntimeError('Unknown output type')



"""
LLVM code generation.

"""
from __future__ import print_function

import contextlib
import functools

import future.moves.collections as collections
from future.builtins import bytes
from llvmlite import ir

import rbc.exception as exc

# HACK: make sure all the AST node types are imported and registered
from . import astnode, expression, external, statement
from . import context

# Constructing AST Nodes
# ======================
#
# AST nodes may be constructed via the make_node function. The advantage of the
# make_node function is one doesn't need to explicitly import all the AST node
# classes.

def make_node(type_name, **kwargs):
    """Construct an AST node given the type name and a set of keyword arguments.

    Returns:
        An :py:class:`.astnode.ASTNode` object.

    Raises:
        InternalCompilerError if type_name does not correspond to a known node.

    """

    try:
        return astnode.make_node(type_name, **kwargs)
    except KeyError:
        raise exc.InternalCompilerError(
            'Unknown AST node: {}'.format(type_name))

# Programs
# ========

@astnode.ast_node
class Program(astnode.ASTNode):
    """A program is the root-level AST node. It provides a single method,
    :py:meth:`.emit`, which returns the LLVM module assembly as a string.

    """
    def emit(self, target, machine):
        """Take an llvm Target and TargetMachine instance representing the
        ultimate target for the emitted code.

        Returns:
            A stirng containing the LLVM module assembly code.

        """
        # Create a new emit context for the program
        ctx = context.EmitContext(target, machine)

        with ctx.emitting_code():
            # Declare all top-level definitions
            for defn in self.definitions:
                defn.declare(ctx)

            # Emit global definitions
            for emittable in self.definitions:
                emittable.emit(ctx)

        return str(ctx.module)



"""
Custom exception subclasses.

"""

class SemanticError(RuntimeError):
    """An error arising from B source which is well-formed in terms of grammar
    but is meaningless in terms of semantics.

    """

class InternalCompilerError(RuntimeError):
    """An error raised when an invalid AST has been created in a manner which
    should not be possible to trigger from B source.

    """



import rbc.compiler as compiler

def test_putchar(check_output):
    check_output('main() { extrn putchar; putchar(65); }', 'A')

def test_putnumb(check_output):
    check_output('main() { extrn putnumb; putnumb(-010); }', '-8')




from __future__ import print_function

from llvmlite import ir

import rbc.exception as exc

from .astnode import ast_node, needs_builder, ASTNode
from .context import (
    address_to_llvm_ptr, llvm_ptr_to_address, create_aligned_global, if_else
)

def get_or_create_string_constant(context, string_bytes):
    """Return an llvm Value which points to a string constant with the
    specified bytes. A terminating *e is appended.

    """
    # Ensure string_bytes is a bytes object
    if not isinstance(string_bytes, bytes):
        string_bytes = bytes(string_bytes)

    # Try to look up string
    cached_ptr = context.string_constants.get(string_bytes)
    if cached_ptr is not None:
        return cached_ptr

    # We need to create one
    str_contents = bytearray(string_bytes)
    str_contents.append(4) # Terminating *e

    # We don't mangle the string constant name so that it won't collide with
    # any B symbols.
    str_name = '__str.{}'.format(len(context.string_constants))
    str_type = ir.ArrayType(ir.IntType(8), len(str_contents))
    str_ptr = create_aligned_global(
        context.module, str_type, context.module.get_unique_name(str_name))
    str_ptr.modifiers = ['align {}'.format(context.bytes_per_word)]

    # The string constant shouldn't be modified and merge it with any other
    # symbols with the same content. They're also local to the module.
    str_ptr.global_constant = True
    str_ptr.unnamed_addr = True
    str_ptr.linkage = 'private'

    # Initialise the variable with the terminated contents
    str_ptr.initializer = ir.Constant(str_type, str_contents)

    # Record the pointer and return
    context.string_constants[string_bytes] = (str_type, str_ptr)

    return str_ptr

# lvalues and rvalues
# ===================
#
# B divides values into two categories: lvalue and rvalue. The essential
# difference is that an lvalue represents the address in memory of some other
# value. Thus each lvalue has an associated rvalue (the value pointed to by the
# address) but each rvalue does not have an associated lvalue. If an rvalue
# *does* have some address associated with it the address may be obtained via
# the "&" or "reference" operator. This address is itself an rvalue. An rvalue
# may not be referenced. Both lvalues and rvalues may be "dereferenced" via the
# "*" operator to yield an lvalue.

class RValue(ASTNode):
    def dereference(self):
        """Return an LValue corresponding to the dereferencing of this value.
        The default implementation returns a DereferencedRValue instance.
        """
        return DereferencedRValue(rvalue=self)

    def reference(self):
        """An RValue cannot be referenced since it does not have an address."""
        raise exc.SemanticError('Cannot reference an rvalue.')

# Our implementation is such that that only lvalues are a dereferenced rvalue or
# an value retrieved from the current scope.

@ast_node
class DereferencedRValue(RValue):
    """A convenience wrapper representing a dereferenced RValue. The reference()
    method simply returns the dereferenced RValue. This value is an LValue in
    that it has a reference() method.

    """
    def reference(self):
        return self.rvalue

    @needs_builder
    def emit(self, context):
        # Convert the rvalue word address into a llvm pointer to a word
        ptr_type = context.word_type.as_pointer()
        rvalue_val = self.rvalue.emit(context)
        word_ptr = address_to_llvm_ptr(context, rvalue_val, ptr_type)

        # Load from pointer
        return context.builder.load(word_ptr)

@ast_node
class ScopeValue(RValue):
    """LValue constructed by dereferencing an address retrieved from the current
    scope."""
    def reference(self):
        return AddressOfScopeValue(name=self.name)

    def emit(self, context):
        try:
            val = context.scope[self.name]
        except KeyError:
            raise exc.SemanticError(
                'Variable not found in scope: {}'.format(self.name))
        return val.emit(context)

# In contrast there are many implementations of rvalues depending on how they're
# calculated.

@ast_node
class AddressOfScopeValue(RValue):
    """RValue corresponding to an address retrieved from the current scope."""
    def emit(self, context):
        try:
            val = context.scope[self.name]
        except KeyError:
            raise exc.SemanticError(
                'Variable not found in scope: {}'.format(self.name))
        return val.reference().emit(context)

@ast_node
class ReferencedLValue(RValue):
    def emit(self, context):
        return self.lvalue.reference().emit(context)

@ast_node
class ConditionalOpValue(RValue):
    @needs_builder
    def emit(self, context):
        with if_else(context, self.cond) as (then, otherwise):
            with then:
                then = self.then.emit(context)
                then_block = context.builder.block
            with otherwise:
                otherwise = self.otherwise.emit(context)
                otherwise_block = context.builder.block

        # We use a phi node to conditionally return the then or otherwise value
        # depending on which block was evaluated.
        value = context.builder.phi(context.word_type)
        value.add_incoming(then, then_block)
        value.add_incoming(otherwise, otherwise_block)
        return value

# Some binary ops map directly into LLVM instructions. They're noted in this
# mapping of operator to instruction name.
_SIMPLE_OPS = {
    '*': 'mul', '/': 'sdiv', '%': 'srem',
    '+': 'add', '-': 'sub',
    '<<': 'shl', '>>': 'lshr',
    '&': 'and_', '^': 'xor', '|': 'or_',
}

# Relational ops need their result casting up to a word. Due to the incredible
# dominance of C, B's successor, the relational operator names haven't changed
# and can be used directly in LLVM IR(!)
_REL_OPS = frozenset(['<', '<=', '>', '>=', '==', '!='])

# Helper functions for binary operators
def _emit_binary_op(context, lhs, op, rhs):
    # No short-cutting in B(!)
    lhs_val, rhs_val = lhs.emit(context), rhs.emit(context)

    if op in _SIMPLE_OPS:
        instr_name = _SIMPLE_OPS[op]
        return getattr(context.builder, instr_name)(lhs_val, rhs_val)

    if op in _REL_OPS:
        cmp_val = context.builder.icmp_signed(op, lhs_val, rhs_val)
        return context.builder.zext(cmp_val, context.word_type)

    raise exc.InternalCompilerError('Unknown binary op: {}'.format(op))

@ast_node
class BinaryOpValue(RValue):
    @needs_builder
    def emit(self, context):
        return _emit_binary_op(context, self.lhs, self.op, self.rhs)

@ast_node
class AssignmentOpValue(RValue):
    def emit(self, context):
        # In an assignment, the lhs is an lvalue and so has an address. Get the
        # address by referencing it and store the rhs to that address. Return
        # the rhs value.
        lhs_addr = self.lhs.reference().emit(context)
        ptr_type = context.word_type.as_pointer()
        lhs_ptr = address_to_llvm_ptr(context, lhs_addr, ptr_type)

        # If the op is anything other than '=', apply the given binary op
        # explicitly and rely on LLVM to perform any optimisation.
        if self.op != '=':
            rhs_val = _emit_binary_op(context, self.lhs, self.op[1:], self.rhs)
            context.builder.store(rhs_val, lhs_ptr)
        else:
            rhs_val = self.rhs.emit(context)
            context.builder.store(rhs_val, lhs_ptr)
        return rhs_val

@ast_node
class LeftUnaryOpValue(RValue):
    def reference(self):
        """The unary op '*' yields an LValue in that &*x is identically x."""
        if self.op != '*':
            return RValue.reference(self)
        return self.rhs

    @needs_builder
    def emit(self, context):
        if self.op == '&':
            return self.rhs.reference().emit(context)
        elif self.op == '*':
            return self.rhs.dereference().emit(context)
        elif self.op == '-':
            rhs = self.rhs.emit(context)
            return context.builder.neg(rhs)
        elif self.op == '~':
            rhs = self.rhs.emit(context)
            return context.builder.not_(rhs)
        elif self.op == '!':
            # Logical not
            rhs = self.rhs.emit(context)
            zero = ir.Constant(context.word_type, 0)
            is_zero = context.builder.icmp_unsigned('==', rhs, zero)
            return context.builder.zext(is_zero, context.word_type)
        elif self.op in ['++', '--']:
            # pre-{inc,dec}rement
            rhs = self.rhs.emit(context)
            rhs_addr_val = self.rhs.reference().emit(context)
            rhs_ptr = address_to_llvm_ptr(
                context, rhs_addr_val, context.word_type.as_pointer())
            one = ir.Constant(context.word_type, 1)
            if self.op == '++':
                val = context.builder.add(rhs, one)
            else:
                val = context.builder.sub(rhs, one)
            context.builder.store(val, rhs_ptr)
            return val

        raise exc.InternalCompilerError('Unknown unary op: {}'.format(self.op))

@ast_node
class RightUnaryOpValue(RValue):
    @needs_builder
    def emit(self, context):
        if self.op in ['++', '--']:
            # post-{inc,dec}rement
            lhs = self.lhs.emit(context)
            lhs_addr_val = self.lhs.reference().emit(context)
            lhs_ptr = address_to_llvm_ptr(
                context, lhs_addr_val, context.word_type.as_pointer())
            one = ir.Constant(context.word_type, 1)
            if self.op == '++':
                val = context.builder.add(lhs, one)
            else:
                val = context.builder.sub(lhs, one)
            context.builder.store(val, lhs_ptr)
            return lhs

        raise exc.InternalCompilerError('Unknown unary op: {}'.format(self.op))

@ast_node
class FunctionCallValue(RValue):
    @needs_builder
    def emit(self, context):
        # Emit function and args expressions to llvm Values
        func_addr_val = self.func.reference().emit(context)
        arg_vals = [arg.emit(context) for arg in self.args]

        # Convert function address into llvm function pointer
        func_type = ir.FunctionType(
            context.word_type, (context.word_type,) * len(arg_vals))
        func_ptr = address_to_llvm_ptr(
            context, func_addr_val, func_type.as_pointer())

        # Call function with arguments
        return context.builder.call(func_ptr, arg_vals)

@ast_node
class BuiltinValue(RValue):
    def emit(self, context):
        if self.name == '__bytes_per_word':
            return ir.Constant(context.word_type, context.bytes_per_word)

        raise exc.InternalCompilerError(
            'Unknown builtin value: {}'.format(self.name))

# Constants
# =========
#
# Constants are RValues whose word value is known at compile time. They share an
# attribute, "value", which is an Python object corresponding to the value of
# the node.

@ast_node
class ConstantIntValue(RValue):
    """An constant integer value."""
    def emit(self, context):
        return ir.Constant(context.word_type, self.value)

@ast_node
class StringConstantValue(RValue):
    @needs_builder
    def emit(self, context):
        # Get a pointer to the string
        str_ptr = get_or_create_string_constant(context, self.value)

        # Return the pointer as an address
        return llvm_ptr_to_address(context, str_ptr)

# Addresses
# =========
#
# LLVM addresses are byte-oriented but B addresses are word-oriented in that,
# given an lvalue "A", "*((&A)+1)" refers to the word which follows A in memory,
# not the word one byte into A. Other languages, such as C, have a type system
# to maintain this whereby each value is known to be either a pointer to a
# particular type or a non-pointer type. B's type system, such as it is, simply
# interprets a value as the type which it is used as. As such, "10[1]" is a
# valid B expression which is interpreted as the 11th word in memory. Addresses
# in B, therefore, are always word addresses.

class LLVMPointerValue(RValue):
    """An RValue corresponding to an llvm pointer type punned to the machine
    word type. The pointer value is divided by the machine word length in bytes
    to yield a value which is word-oriented.

    It is a Bad Idea (TM) to pass this AST node a pointer which is not
    word-aligned. There is no runtime check for alignment. So, "caveat caller".

    This class is not marked as an AST node and so cannot be constructed via the
    make_node() function. This reflects the fact that this is an "internal" node
    not intended for construction by the semantics.

    """
    @needs_builder
    def emit(self, context):
        return llvm_ptr_to_address(context, self.value)



def test_not_operator(check_expr):
    check_expr('!5', '0')
    check_expr('!-1', '0')
    check_expr('!1', '0')
    check_expr('!0', '1')

def test_minus(check_expr):
    check_expr('-5', '-5')
    check_expr('-0', '0')

def test_reference(check_output):
    check_output('''
        main() {
            extrn v, putnumb;
            putnumb(*(v+1));
        }
        v[] 1, 2, 3;
    ''', 2)

def test_dereference(check_output):
    check_output('''
        main() {
            extrn putnumb;
            auto i;
            i = 56;
            putataddr(&i);
            putnumb(i);
        }
        putataddr(addr) { extrn putnumb; putnumb(*addr); *addr=0; }
    ''', '560')

def test_preinc(check_output):
    check_output('''
        main() { extrn putnumb; auto i; i=4; putnumb(++i); putnumb(i); }
    ''', '55')

def test_predec(check_output):
    check_output('''
        main() { extrn putnumb; auto i; i=4; putnumb(--i); putnumb(i); }
    ''', '33')

def test_postinc(check_output):
    check_output('''
        main() { extrn putnumb; auto i; i=4; putnumb(i++); putnumb(i); }
    ''', '45')

def test_postdec(check_output):
    check_output('''
        main() { extrn putnumb; auto i; i=4; putnumb(i--); putnumb(i); }
    ''', '43')

def test_dereference_via_flip(check_output):
    check_output('''
        flip(x, y) { auto t; t = *y; *y = *x; *x = t; }
        main() {
            extrn putnumb;
            auto a, b;
            a = 10; b = 20;
            putnumb(a); putnumb(b);
            flip(&a, &b);
            putnumb(a); putnumb(b);
        }
    ''', '10202010')



def test_char(check_output):
    check_output('''
        main() {
            extrn char, putchar, s;
            auto i;
            i = 0; while(i <=4) {
                putchar('>'); putchar(char(s, i)); putchar('<');
                ++i;
            }
        }
        s "hello";
    ''', '>h<>e<>l<>l<>o<')

def test_lchar(check_output):
    check_output('''
        main() {
            extrn lchar, putstr;
            auto s[3];
            lchar(s, 0, 'X');
            lchar(s, 1, 'Y');
            lchar(s, 2, 'Z');
            lchar(s, 3, '*e');
            putstr(s);
        }
    ''', 'XYZ')

def test_printn(check_output):
    check_output('''
        main() {
            extrn printn, putchar;
            printn(42, 2); putchar('*n');
            printn(42, 8); putchar('*n');
            printn(42, 10); putchar('*n');
        }
    ''', '101010\n52\n42\n')



"""
Abstract syntax tree nodes capable of emitting LLVM instructions.

"""
import functools

import rbc.exception as exc

# Constructing AST Nodes
# ======================
#
# AST nodes may be constructed directly via the class object or indirectly via
# the make_node function. The advantage of the make_node function is one doesn't
# need to explicitly import all the AST node class names.

# A map of node names -> node classes used by make_node.
_NODE_CLASSES = {}

def make_node(type_name, **kwargs):
    """Construct an AST node given the type name and a set of keyword arguments.

    Returns:
        An ASTNode object.

    Raises:
        KeyError if type_name does not correspond to a known node.

    """
    return _NODE_CLASSES[type_name](**kwargs)

def ast_node(cls):
    """Class decorator which registers the AST node in the _NODE_CLASSES map."""
    _NODE_CLASSES[cls.__name__] = cls
    return cls

# AST Nodes
# =========
#
# AST nodes fundamentally store key/value associations much like dicts. Unlike
# dicts AST nodes don't support iterating over the values and instead expose the
# values directly as attributes.

class ASTNode(object):
    """An AST node with parameters directly accessible as attributes."""
    def __init__(self, **kwargs):
        # Set attributes on ourself directly from the keyword args.
        for k, v in kwargs.items():
            setattr(self, k, v)

# Emittable nodes
# ===============
#
# Most AST nodes are "emittables" in that they have an emit() method which takes
# the current context and returns an llvm Value corresponding to that node. Some
# nodes do not have a corresponding llvm Value and so return None from emit().
# The Program node returns the entire LLVM module assembly as a string value.

def needs_builder(emit):
    """A decorator for emit() methods which make use of the context's builder
    attribute. If the builder attribute is None, an InternalCompilerError is
    raised.

    """
    @functools.wraps(emit)
    def _wrapped_emit(self, context):
        if context.builder is None:
            raise exc.InternalCompilerError('AST node requires builder.')
        return emit(self, context)
    return _wrapped_emit



"""
Backports of Python functionality.

"""
import shutil as _shutil
import sys as _sys
import tempfile as _tempfile

__all__ = ['TemporaryDirectory']

# Backported from Python 3.2.
class _TemporaryDirectoryBackport(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:
        with TemporaryDirectory() as tmpdir:
            ...
    Upon exiting the context, the directory and everthing contained
    in it are removed.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        self.name = None # Handle mkdtemp raising an exception
        self.name = _tempfile.mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self):
        if self.name and not self._closed:
            try:
                _shutil.rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                _sys.stderr.write("ERROR: {!r} while cleaning up {!r}\n".format(
                    ex, self,))
                return
            self._closed = True

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        self.cleanup()

# Try standard tempfile implementation in preference to backported one
try:
    TemporaryDirectory = _tempfile.TemporaryDirectory
except AttributeError:
    TemporaryDirectory = _TemporaryDirectoryBackport



"""
Setup configuration for installation via pip, easy_install, etc.

"""
import sys
from setuptools import setup, find_packages

# Python 3.4+ comes with enum.Enum. For prior Pythons, we can install the enum34
# package from PyPI.
if sys.hexversion < 0x030400F0:
    enum_requires = ['enum34']
else:
    enum_requires = []

# The find_packages function does a lot of the heavy lifting for us w.r.t.
# discovering any Python packages we ship.
setup(
    name='rbc',
    version='0.2.0',
    description='Example B compiler written with LLVM',
    url='https://github.com/rjw57/rbc',
    author='Rich Wareham',
    author_email='rich.rbc@richwareham.com',
    license='MIT',
    packages=find_packages(),

    # PyPI packages required for the *installation* and usual running of the
    # tools.
    install_requires=[
        'docopt',
        'future',
        'grako',
        'llvmlite',
        'whichcraft',
    ] + enum_requires,

    entry_points={
        'console_scripts': [
            'rbc=rbc:main',
        ],
    },

    package_data={
        'rbc': ['*.c', '*.b'],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    keywords='compiler B llvm example',
)



"""
Usage:
    dumpast.py (--json | --dot) [-p TERM] [<file>]

Options:
    -p TERM     Start parsing from the given term. [default: program]
    --json      Output in JSON format.
    --dot       Output in Graphviz format.

If <file> is omitted, input is read from standard input.

"""
from __future__ import unicode_literals

import collections
import json
import sys

from future.builtins import bytes

import docopt

from rbc.parser import BParser
from rbc.semantics import BSemantics

# A debugging tool for dumping ASTs in JSON or Graphviz format.

def make_ordered_dict_ast_node(type_, **kwargs):
    """Return a collections.OrderedDict with the type as the first value
    with key "_type".
    """
    od = collections.OrderedDict(_type=type_)
    od.update(**kwargs)
    return od

class ASTJSONEncoder(json.JSONEncoder):
    """An extended JSON encoder which can handle the types in AST nodes."""
    # pylint: disable=method-hidden
    def default(self, obj):
        if isinstance(obj, bytes):
            return repr(obj)
        return json.JSONEncoder.default(self, obj)

def escape_label(label):
    label = label.replace('\\', '\\\\')
    label = label.replace('"', '\\"')
    label = label.replace('<', '\\<')
    label = label.replace('>', '\\>')
    label = label.replace('|', '\\|')
    label = label.replace('{', '\\{')
    label = label.replace('}', '\\}')
    return label

class GraphvizAST(object):
    Node = collections.namedtuple('Node', 'name fields')

    def __init__(self):
        # A list of Node instances
        self._nodes = []

        # A set of Node pairs
        self._edges = set()

    def make_node(self, type_, **kwargs):
        # Each node has a unique id
        name = 'Node{}'.format(len(self._nodes))
        fields = [type_]

        items = list(kwargs.items())

        # Sort the items so that the order is at least consistent
        items = sorted(items, key=lambda item: item[0])

        extended_items = []
        for k, v in items:
            if isinstance(v, list) and len(v) > 0:
                extended_items.extend(('{}[{}]'.format(k, vidx), velem)
                                      for vidx, velem in enumerate(v))
            else:
                extended_items.append((k, v))

        for k, v in extended_items:
            if isinstance(v, GraphvizAST.Node):
                self._edges.add((
                    '{}:<f{}>'.format(name, len(fields)),
                    '{}:<f0>'.format(v.name)
                ))
                fields.append(k)
            else:
                fields.append('{}: {}'.format(k, v))

        # Return the new node
        node = GraphvizAST.Node(name=name, fields=tuple(fields))
        self._nodes.append(node)
        return node

    def emit(self, fobj):
        fobj.write('digraph {\n')
        fobj.write('graph [ rankdir="LR" ]; node [ shape=record ];\n')

        for node in self._nodes:
            label = ' | '.join('<f{}> {}'.format(idx, escape_label(label))
                               for idx, label in enumerate(node.fields))
            fobj.write('{} [ label="{}" ] ;\n'.format(node.name, label))

        for a, b in self._edges:
            fobj.write('{} -> {} ;\n'.format(a, b))

        fobj.write('\n}\n')

def main():
    opts = docopt.docopt(__doc__)
    if opts['<file>'] is not None:
        with open(opts['<file>']) as fobj:
            source = fobj.read()
    else:
        source = sys.stdin.read()

    if opts['--dot']:
        ast = GraphvizAST()
        out_format = 'dot'
        make_node = ast.make_node
    elif opts['--json']:
        out_format = 'json'
        make_node = make_ordered_dict_ast_node
    else:
        raise RuntimeError('No output format in options')

    node = BParser().parse(source, opts['-p'], semantics=BSemantics(make_node))

    if out_format == 'json':
        encoder = ASTJSONEncoder(indent=2)
        for chunk in encoder.iterencode(node):
            sys.stdout.write(chunk)
        sys.stdout.write('\n')
    else:
        ast.emit(sys.stdout)

if __name__ == '__main__':
    main()



"""
High-level interface to the B compiler.

"""
import os
import subprocess

import llvmlite.binding as llvm
import pkg_resources
import whichcraft

import rbc.codegen as codegen

from rbc.parser import BParser
from rbc.semantics import BSemantics
from rbc._backport import TemporaryDirectory

# pylint: disable=assignment-from-no-return
_LIBB_C_SOURCE_FILE = pkg_resources.resource_filename(__name__, 'libb.c')
_LIBB_B_SOURCE_FILE = pkg_resources.resource_filename(__name__, 'libb.b')

def _ensure_llvm():
    """Ensure that LLVM has been initialised."""
    if _ensure_llvm.was_initialized:
        return

    llvm.initialize()
    llvm.initialize_native_target()
    llvm.initialize_native_asmprinter()
    _ensure_llvm.was_initialized = True

_ensure_llvm.was_initialized = False


class CompilerOptions(object):
    """There are many options which affect the behaviour of the compiler. They
    are collected into this class for easy transport.

    The target and machine attributes are given default values based on the host
    machine running the compiler. The default optimisation level is 1.

    IMPORTANT: Make sure that LLVM and in particular the native target has been
    initialised via the llvmlite.binding.initialize...() functions before
    constructing an instance of this object.

    Attributes:
        target: The llvm.Target which is the target of compilation.
        machine: The llvm.TargetMachine which is the target of compilation.
        opt_level: The optimisation level from 0 (no optimisation) to 3 (full
                   optimisation.)

    """
    def __init__(self):
        _ensure_llvm()
        self.target = llvm.Target.from_default_triple()
        self.machine = self.target.create_target_machine(codemodel='default')
        self.opt_level = 1

def compile_b_source(source, options):
    """The B front end converts B source code into a LLVM module. No significant
    optimisation is performed.

    Args:
        source (str): B source code as a string
        options (CompilerOptions): compiler options

    Returns:
        A string with the LLVM assembly code for an unoptimised module
        corresponding to the input source.

    """
    # Set parser semantics and go forth and parse.
    program = BParser().parse(source, 'program',
                              semantics=BSemantics(codegen.make_node))

    # Emit LLVM assembly for the correct target.
    module_str = program.emit(options.target, options.machine)

    # Return the string representation of the module.
    return module_str

def optimize_module(module_assembly, options):
    """Verify and optimise the passed LLVM module assembly.

    Args:
        module_assembly (str): LLVM module assembly
        options (CompilerOptions): options for the compiler

    Returns:
        A llvmlite.binding.ModuleRef for the verified and optimised module.

    """
    _ensure_llvm()

    # Parse LLVM module assembly
    module = llvm.parse_assembly(module_assembly)
    module.verify()

    # Create optimiser pass manager
    pass_manager = llvm.ModulePassManager()

    # Populate with target passes
    options.machine.target_data.add_pass(pass_manager)

    # Populate with optimisation passes
    pass_manager_builder = llvm.PassManagerBuilder()
    pass_manager_builder.opt_level = options.opt_level
    pass_manager_builder.populate(pass_manager)

    # Run optimiser
    pass_manager.run(module)

    return module

class CompilationEnvironment(object):
    """
    Detect compiler tools available in the environment.

    Some parts of ``rbc`` call out to external compiler tools. This class
    centralises the automatic discovery of these tools. Custom environments may
    be created by creating an instance of this class and setting attributes
    manually.

    Attributes:
        gcc: path to the GCC compiler binary or None if no GCC present
        cppflags: list of C pre-processor flags
        cflags: list of C compiler flags
        ldflags: list of linker flags

    """
    def __init__(self):
        self.gcc = whichcraft.which('gcc')
        self.cflags = ['-std=gnu99']
        self.cppflags = []
        self.ldflags = []

    def compile_c_source(self, obj_filename, c_filename):
        subprocess.check_call(
            [self.gcc] + self.cppflags + self.cflags +
            ['-c', '-o', obj_filename, c_filename])

    def link_objects(self, output_filename, obj_filenames):
        subprocess.check_call(
            [self.gcc] + self.ldflags +
            ['-o', output_filename] + obj_filenames)

_DEFAULT_ENVIRONMENT = CompilationEnvironment()

def compile_b_to_native_object(obj_filename, b_filename, options):
    """Convenience function to compile an on-disk B file to a native object.

    Args:
        obj_filename (str): file to write object code to
        b_filename (str): file containing B source
        options (CompilerOptions): compiler options to use

    """
    with open(b_filename) as fobj:
        source = fobj.read()

    module_asm = compile_b_source(source, options)
    module = optimize_module(module_asm, options)
    module.name = os.path.basename(b_filename)

    with open(obj_filename, 'wb') as fobj:
        fobj.write(options.machine.emit_object(module))

def compile_and_link(output, source_files, options=None,
                     env=_DEFAULT_ENVIRONMENT):
    """Compile and link source files into an output file. Uses GCC for the heavy
    lifting. This will implicitly link in the B standard library.

    Input files may be anything GCC accepts along with B source files.

    If no compiler options are used, a new CompilerOptions object is
    constructed.

    Note: the passed compiler options *only* affect the B compiler. Use the
    'cflags', 'ldflags' and 'cppflags' attributes in the compilation
    environment.

    Args:
        output (str): path to output file
        source_files (sequence): paths of input files
        options (CompilerOptions): compiler options
        env (CompilationEnvironment): specify custom compiler environment

    """
    options = options if options is not None else CompilerOptions()

    with TemporaryDirectory() as tmp_dir:
        libb1_obj = os.path.join(tmp_dir, 'libb1.o')
        env.compile_c_source(libb1_obj, _LIBB_C_SOURCE_FILE)
        libb2_obj = os.path.join(tmp_dir, 'libb2.o')
        compile_b_to_native_object(libb2_obj, _LIBB_B_SOURCE_FILE, options)
        compiled_source_files = [libb1_obj, libb2_obj]
        for file_idx, source_file in enumerate(source_files):
            out_file = os.path.join(tmp_dir, 'tmp{}.o'.format(file_idx))
            _, ext = os.path.splitext(source_file)
            if ext == '.b':
                compile_b_to_native_object(out_file, source_file, options)
                compiled_source_files.append(out_file)
            elif ext == '.c':
                env.compile_c_source(out_file, source_file)
                compiled_source_files.append(out_file)
            else:
                compiled_source_files.append(source_file)
        env.link_objects(output, compiled_source_files)



def test_simple_goto(check_output):
    check_output('''
        main() {
            auto i;
            extrn putchar;
            putchar('X');
            i = 0;
        loop:
            putchar('.');
            if(++i != 5) goto loop;
            putchar('Y');
        }
    ''', 'X.....Y')

def test_goto_future(check_output):
    check_output('''
        main() {
            auto i;
            extrn putchar;
            putchar('X');
            i = 0;
        loop:
            putchar('.');
            if(++i == 5) goto exit;
            goto loop;
        exit:
            putchar('Z');
        }
    ''', 'X.....Z')



def test_count_down(check_output):
    check_output('''
        countdown(num) {
            extrn putnumb;
            while(num > 0) { putnumb(num); num =- 1; }
        }
        main() {
            countdown(5);
        }
    ''', '54321')

def test_null_statement(check_output):
    check_output('main(){while(0);}', '')

def test_break(check_output):
    check_output('''
        main() {
            auto i;
            extrn putchar;
            i = 0; while(i < 100) {
                putchar('.');
                if(++i >= 4)
                    break;
            }
        }
    ''', '....')

def test_break_is_no_op_outside_while(check_output):
    check_output('''
        main() { extrn putchar; break; putchar('X'); }
    ''', 'X')







def test_simple_if(check_output):
    check_output('''
        main(){
            extrn putnumb;
            if(0) putnumb(1);
            if(1) putnumb(2);
            if(-1) putnumb(3);
            if(42) putnumb(4);
            if(-42) putnumb(5);
        }
    ''', '2345')

def test_if_based_return(check_output):
    check_output('''
        max(a, b) {
            if(a > b) return(a);
            return(b);
        }
        main() {
            extrn putnumb;
            putnumb(max(9, -7));
            putnumb(max(3, 5));
        }
    ''', '95')

def test_simple_if_else(check_output):
    check_output('''
        putargmax(a,b) {
            extrn putnumb;
            if(a>=b) putnumb(0); else putnumb(1);
        }
        main() {
            putargmax(10, 10);
            putargmax(10, 11);
            putargmax(5, 4);
        }
    ''', '010')

def test_else_binds_to_inner_if(check_output):
    check_output('''
        posnegzero(a) {
            if(a != 0) if(a > 0) return (1); else return(-1);
            return(0);
        }
        main() {
            extrn putnumb;
            putnumb(posnegzero(-2));
            putnumb(posnegzero(-1));
            putnumb(posnegzero( 0));
            putnumb(posnegzero( 1));
            putnumb(posnegzero( 2));
        }
    ''', '-1-1011')

